/**
 * @author Olgierd Humeńczuk olgierdh@gmail.com
 * @file main
 */

#include <iostream>
#include <stdlib.h>
#include <stdexcept>
#include <vector>
#include <math.h>
#include <cassert>

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <SDL.h>
#else
#include <SDL/SDL.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#undef main

#include "math.hpp"
#include "mesh.hpp"
#include "renderer.hpp"
const double EPSILON = 0.001;

/**
 * @brief Funkcja przetwarza zdarzenia pochodzace z kontrolerów takich jak:
 *          * klawiatura
 *          * mysz komputerowa
 *
 * @return true jeżeli aplikacja ma pozostać otwarta, false wpp
 */
bool processInput( void )
{
    // zmienna pomocnicza
    SDL_Event event;

    // czy są jakieś czekające na przetworzenie zdarzenia ?
    if ( SDL_PollEvent( &event ) )
    {
        // sprawdzenie typu zdarzenia
        switch( event.type )
        {
            // wyjście z aplikacji
            case SDL_QUIT:
                return false;

            // użytkownik nacisnął któryś z klawiszy na klawiaturze
            case SDL_KEYDOWN:
                // jeżeli nacisnał klawisz eskape zamykamy aplikacje
                if( event.key.keysym.sym == SDLK_ESCAPE )
                {
                    return false;
                }
                break;
            case SDL_MOUSEMOTION:
                SDL_GetMouseState( &mousePosition.x, &mousePosition.y );
                break;
        }
    }
    return true;
}

/**
 * @brief Funkcja odpowiadająca za inicjalizację biblioteki SDL
 *          oraz utworzenie okna z kontekstem OpenGL.
 * @param scrWidth szerokosc ekranu
 * @param scrHeight wysokosc ekranu
 *
 * @throw std::runtime_error jeżeli nastąpi jakikolwiek błąd.
 */
void initSDL( int scrWidth, int scrHeight )
{
    // zaincjalizowanie części biblioteki odpowiadającej za wyświetlanie
    if( SDL_Init( SDL_INIT_VIDEO ) < 0 )
    {
        throw std::runtime_error( "Can't initialize SDL\n" );
    }

    // ustawienie ilości bitów dla każdego z kolorów
    SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
    SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
    SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );

    // ustawienie ilości bitów dla specjalnych buforów
    SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 24 );
    SDL_GL_SetAttribute( SDL_GL_STENCIL_SIZE, 8 );

    // włączenie podwójnego buforowania
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, true );
    SDL_GL_SetAttribute( SDL_GL_SWAP_CONTROL, 1 );

    // stworzenie okna aplikacji z kontekstem OpenGL'a
    if ( SDL_SetVideoMode( scrWidth, scrHeight, 24, SDL_OPENGL ) == 0 )
    {
        throw std::runtime_error(
            std::string( "Unable to set video mode: " ).append(
            	std::string( SDL_GetError() ) ) );
    }
}

/**
 * @brief Główna metoda programu ( entry point )
 * @param argc ilość argumentów przekazanych do programu
 * @param argv tablica wartości argumentów
 */
int main( const int /*argc*/, const char** /*argv*/ )
{
	 const int screenX = 800;
	 const int screenY = 600;

    try
    {
        initSDL( screenX, screenY );
        initOpenGL( screenX, screenY );

        vertexes cubeVertexes = createSphere( 2.0f, 9, 9 );

        glEnableClientState( GL_VERTEX_ARRAY );
        glEnableClientState( GL_NORMAL_ARRAY );

        float angle = .0f;

        // główna pętla
        while( processInput() )
        {
            glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
            glMatrixMode( GL_MODELVIEW );
            glLoadIdentity();

            glPushMatrix();
            gluLookAt( 0.0f, 0.0f, 20.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );

            {
                //dane światła
                GLfloat position[] = { 0.0f, 10.0f, 10.0f, 1.0f };
                GLfloat ambient[] = { 0.0f, 0.0f, 0.0f, 1.0f };
                GLfloat diffuse[] = { 0.8f, 0.8f, 0.8f, 1.0f };
                GLfloat specular[] = { 0.8f, 0.8f, 0.8f, 1.0f };

                setPointLight( GL_LIGHT0, position, ambient, diffuse, specular );
            }

            glPushMatrix();
            glTranslatef(
                transformPos( static_cast< GLfloat >( mousePosition.x ),
			      static_cast< GLfloat >( screenX ), -5.5f, 5.5f ),
                0.0f, 0.0f );

            glRotatef( angle, 1.0f, .0f, .0f );

            //prędkość obrotu
            angle += 0.4f;

            glVertexPointer( 3, GL_FLOAT, sizeof( vertexes::value_type ), &cubeVertexes[ 0 ] );
            glNormalPointer( GL_FLOAT, sizeof( vertexes::value_type ), &cubeVertexes[ 0 ].normal );

            {
                //dane materialu
                GLfloat ambient[] = { 0.0f, 0.0f, 0.0f, 1.0f };
                GLfloat diffuse[] = { 1.0f, 0.0f, 0.0f, 1.0f };
                GLfloat specular[] = { 0.2f, 0.2f, 0.2f, 1.0f };

                setMaterial( ambient, diffuse, specular, 0.9f );
            }

            glDrawArrays( GL_QUADS, 0, cubeVertexes.size() );

            glPopMatrix();
            glPopMatrix();

            SDL_GL_SwapBuffers();
        }
    }
    catch( const std::exception& e )
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
    }

    SDL_Quit();
    return 0;
}
